/*
 *	SynCom -- Decoder
 *
 *	(c) 1999 Martin Mares <mj@ucw.cz>
 *
 *	Can be freely distributed and used under the terms of the GNU GPL.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <fcntl.h>
#include <unistd.h>

#include "syncom.h"

static FILE *src_fh, *dst_fh;

static int num_syms;
static struct node **symbols;

static void
read_symbols(void)
{
  int i;
  char buf[256], *d;
  struct node *n;

#ifdef FAKE_FORMAT
  fscanf(src_fh, "%d\n", &num_syms);
#else
  fread(&num_syms, sizeof(num_syms), 1, src_fh);
  huff_init(src_fh, 256);
#endif
  symbols = xmalloc(sizeof(struct node *) * num_syms);
  for(i=0; i<num_syms; i++)
    {
#ifdef FAKE_FORMAT
      char *c;
      fgets(buf, sizeof(buf)-1, src_fh);
      c = strchr(buf, '\n');
      if (!c)
	die("Malformed symbol table");
      *c = 0;
#else
      char *c = buf;
      int j;
      while (j = huff_decode())
	*c++ = j;
      *c = 0;
#endif
      n = symbols[i] = xmalloc(sizeof(struct node) + strlen(buf) + 1);
      d = (char *)(symbols[i] + 1);
      strcpy(d, buf);
      n->children = n->sibling = NULL;
      n->leaf = 1;
      n->u.t = d;
    }
}

static struct node *tos = NULL;

static inline void
push(struct node *n)
{
  n->sibling = tos;
  tos = n;
}

static inline struct node *
pop(void)
{
  struct node *n = tos;

  if (!n)
    die("Stack underflow");
  tos = n->sibling;
  return n;
}

static void
action(int rule, char *cmd, ...)
{
  va_list args;
  struct node *n, *m;

  va_start(args, cmd);
  n = new_node(rule, NULL);
  while (*cmd)
    {
      switch (*cmd++)
	{
	case '@':
	  m = pop();
	  break;
	case '#':
	  m = new_node(0, NULL);
	  m->u.t = va_arg(args, char *);
	  m->leaf = 1;
	  break;
	default:
	  die("Internal error: invalid rule: %c", cmd[-1]);
	}
      m->sibling = n->children;
      n->children = m;
    }
  push(n);
  va_end(args);
}

static void
do_rule(int rule)
{
  switch (rule)
    {
#define RULE(x,y) case x: action y; break;
#include "rules.h"
    default:
      die("Unknown rule %d", rule);
    }
}

#ifdef FAKE_FORMAT

static struct node *
read_tree(void)
{
  for(;;)
    {
      int i = fgetc(src_fh);
      int j;

      if (i == EOF)
	die("Unexpected end of file");
      switch (i)
	{
	case '\n':
	  continue;
	case 'E':
	  if (tos->sibling)
	    die("Stack not empty at the end");
	  return tos;
	case 'P':
	  fscanf(src_fh, "%d", &j);
	  if (j < 0 || j >= num_syms)
	    die("Symbol out of range");
	  push(new_node(1, symbols[j], NULL));
	  break;
	case 'R':
	  fscanf(src_fh, "%d", &j);
	  if (j < 0 || j >= NUM_RULES)
	    die("Rule out of range");
	  do_rule(j);
	  break;
	default:
	  die("Unrecognized command %c", i);
	}
    }
}

#else

static struct node *
read_tree(void)
{
  huff_init(src_fh, num_syms + NUM_RULES);
  for(;;)
    {
      int i = huff_decode();

      if (i == num_syms)
	{
	  if (tos->sibling)
	    die("Stack not empty at the end");
	  return tos;
	}
      else if (i < num_syms)
	push(new_node(1, symbols[i], NULL));
      else
	do_rule(i - num_syms);
    }
}

#endif

static int column = 0;

static void
output_tree(struct node *n)
{
  if (!n->leaf)
    {
      for(n=n->children; n; n=n->sibling)
	output_tree(n);
    }
  else
    {
      char *c = n->u.t;
      int len = strlen(c);
      if (column)
	{
	  if (column + 1 + len > 80)
	    {
	      fputc('\n', dst_fh);
	      column = 0;
	    }
	  else
	    {
	      fputc(' ', dst_fh);
	      column++;
	    }
	}
      fputs(c, dst_fh);
      column += len;
    }
}

int
main(int argc, char **argv)
{
  struct node *root;

  if (argc != 3)
    die("Usage: decode <from> <to>");

  src_fh = fopen(argv[1], "r");
  if (!src_fh)
    die("Unable to open source file: %m");
  read_symbols();
  root = read_tree();
  fclose(src_fh);

  dst_fh = fopen(argv[2], "w");
  if (!dst_fh)
    die("Unable to open destination file: %m");
  output_tree(root);
  fputc('\n', dst_fh);
  fclose(dst_fh);

  return 0;
}
