/*
 *	SynCom -- Encoder
 *
 *	(c) 1999 Martin Mares <mj@ucw.cz>
 *
 *	Can be freely distributed and used under the terms of the GNU GPL.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <fcntl.h>
#include <unistd.h>
#include <ctype.h>

#include "syncom.h"
#include "parse.tab.h"

#define RULE(x,y)
#include "rules.h"

int lino;				/* Line number in source file */
struct node *root;

static int src_fd;
static FILE *dst_fh;

void
error(char *msg, ...)
{
  va_list args;

  va_start(args, msg);
  fprintf(stderr, "Line %d: ", lino);
  vfprintf(stderr, msg, args);
  fputc('\n', stderr);
  va_end(args);
  exit(1);
}

int
read_input(char *buf, int max)
{
  return read(src_fd, buf, max);
}

#define HASH_SIZE 4096
static struct keyword *kwd_hash[HASH_SIZE];
static struct keyword *first_kwd, **last_kwd = &first_kwd;

static inline unsigned int
khash(char *w)
{
  unsigned int x = 0;

  while (*w)
    x = x*37 + 
#ifdef CASE_INSENSITIVE
      toupper(*w++)
#else
      *w++
#endif
      ;
  return x & (HASH_SIZE - 1);
}

struct keyword *
intern(char *kwd, int class)
{
  struct keyword *w;
  unsigned int h = khash(kwd);

  for(w=kwd_hash[h]; w && (
#ifdef CASE_INSENSITIVE
      strcasecmp
#else
      strcmp
#endif
	(w->w, kwd) || w->class != class); w=w->sibling)
     ;
  if (w)
    return w;
  w = xmalloc(sizeof(struct keyword) + strlen(kwd));
  *last_kwd = w;
  last_kwd = &w->next;
  w->next = NULL;
  w->sibling = kwd_hash[h];
  kwd_hash[h] = w;
  w->class = class;
  w->id = 0;
  strcpy(w->w, kwd);
  return w;
}

static void
init_tokens(void)
{
  struct { char *w; int i; } tok[] = {
#include "kwd.h"
  };
  unsigned int i;

  for(i=0; i<sizeof(tok)/sizeof(tok[0]); i++)
    {
      struct keyword *w = intern(tok[i].w, 0);
      w->id = tok[i].i;
    }
  first_kwd = NULL;
  last_kwd = &first_kwd;
}

#ifdef FAKE_FORMAT

static void
send_start(void)
{
}

static void
send_symbols(void)
{
  struct keyword *w;
  int cnt = 0;

  for(w=first_kwd; w; w=w->next)
    w->id = cnt++;
  fprintf(dst_fh, "%d\n", cnt);
  for(w=first_kwd; w; w=w->next)
    fprintf(dst_fh, "%s\n", w->w);
}

static void
send_tree(struct node *n)
{
  if (!n->leaf)
    {
      struct node *m;
      for(m=n->children; m; m=m->sibling)
	send_tree(m);
      if (n->u.id)
	fprintf(dst_fh, "R%d\n", n->u.id);
    }
  else
    fprintf(dst_fh, "P%d\n", n->u.w->id);
}

static void
send_end(void)
{
  fprintf(dst_fh, "E\n");
}

#else

static void
send_start(void)
{
}

static int num_syms;

static void
send_symbols(void)
{
  struct keyword *w;

  for(w=first_kwd; w; w=w->next)
    w->id = num_syms++;
  fwrite(&num_syms, sizeof(num_syms), 1, dst_fh);
  huff_init(dst_fh, 256);
  for(w=first_kwd; w; w=w->next)
    {
      char *s;
      for(s=w->w; *s; s++)
	huff_encode(*s);
      huff_encode(0);
    }
  huff_end_write();
  huff_init(dst_fh, num_syms + NUM_RULES);
}

static void
send_tree(struct node *n)
{
  if (!n->leaf)
    {
      struct node *m;
      for(m=n->children; m; m=m->sibling)
	send_tree(m);
      if (n->u.id)
	huff_encode(n->u.id + num_syms);
    }
  else
    huff_encode(n->u.w->id);
}

static void
send_end(void)
{
  huff_encode(num_syms);
  huff_end_write();
}

#endif

int
main(int argc, char **argv)
{
  if (argc != 3)
    die("Usage: encode <from> <to>");
  src_fd = open(argv[1], O_RDONLY);
  if (src_fd < 0)
    die("Unable to open source file: %m");

  lino = 1;
  init_tokens();
  yyparse();
#ifdef FAKE_FORMAT
  dump_tree(root);
#endif
  close(src_fd);

  dst_fh = fopen(argv[2], "w");
  if (!dst_fh)
    die("Unable to open destination file: %m");
  send_start();
  send_symbols();
  send_tree(root);
  send_end();
  fclose(dst_fh);

  return 0;
}
