/*
 *	Dynamic Huffman Coding and Decoding
 *
 *	(c) 1999 Martin Mares <mj@ucw.cz>
 *
 *	Can be freely distributed and used under the terms of the GNU GPL.
 *
 *	This program is based on the original algorithm by Faller,
 *	Gallager and Knuth as described by J. S. Vitter in
 *
 *		Design and Analysis of Dynamic Huffman Codes
 *		(http://www.duke.cs.edu/~jsv/Papers/Vit87.jacm.pdf)
 *
 *	This is still a trivial implementation...
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "syncom.h"

static int num_sym;			/* Number of symbols */
static int bits_per_sym;		/* Bits per symbol (used after ESC) */
static FILE *hfile;

/*
 *	Bit I/O
 */

static int tmpchar, bitmask;

static inline int
get_bit(void)
{
  int t;

  if (!bitmask)
    {
      tmpchar = fgetc(hfile);
      if (tmpchar == EOF)
	die("Unexpected end of file");
      bitmask = 0x80;
    }
  t = !!(tmpchar & bitmask);
  bitmask >>= 1;
  return t;
}

static inline void
put_bit(int bit)
{
  if (!bitmask)
    {
      tmpchar = 0;
      bitmask = 0x80;
    }
  if (bit)
    tmpchar |= bitmask;
  bitmask >>= 1;
  if (!bitmask)
    fputc(tmpchar, hfile);
}

void
huff_end_write(void)
{
  while (bitmask)
    put_bit(0);
}

/*
 *	Tree operations
 */

#define ESC -1
#define INTERNAL -2

struct hnode {
  struct hnode *parent, *son[2];	/* Tree links */
  struct hnode *next, *prev;		/* Next and previous node in our numbering */
  int symbol;				/* Symbol corresponding to this node or ESC or INTERNAL */
  int cnt;				/* Occurence counter */
  struct hnode *aux;			/* Aux pointer used during operations to represent stack */
};

static struct hnode esc_node;
static struct hnode *hroot;
static struct hnode **hindex;

void
huff_init(FILE *f, int n)
{
  hfile = f;
  num_sym = n;
  for(bits_per_sym=1; (1 << bits_per_sym) < num_sym; bits_per_sym++);
  hindex = xmalloc(sizeof(struct hindex *) * (num_sym+1));
  bzero(hindex, sizeof(struct hindex *) * (num_sym+1));
  hindex++;

  bzero(&esc_node, sizeof(esc_node));
  hroot = &esc_node;
  esc_node.next = esc_node.prev = &esc_node;
  esc_node.symbol = ESC;
  hindex[ESC] = &esc_node;

  tmpchar = 0;
  bitmask = 0;
}

#ifdef DEBUG

static void
hdump2(struct hnode *h, int level, struct hnode *e)
{
  int i;

  for(i=0; i<level; i++)
    fputc(' ', stderr);
  if (h->parent != e)
    die("BAD PTR");
  if (h->symbol != INTERNAL)
    fprintf(stderr, "<%d> [%d]\n", h->symbol, h->cnt);
  else
    {
      fprintf(stderr, "INT [%d]\n", h->cnt);
      hdump2(h->son[0], level+2, h);
      hdump2(h->son[1], level+2, h);
    }
}

static void
hdump(void)
{
  struct hnode *h, *e;

  hdump2(hroot, 0, NULL);
  fprintf(stderr, "List: ");
  h = &esc_node;
  e = esc_node.prev;
  do
    {
      fprintf(stderr, "%d[%d] ", h->symbol, h->cnt);
      if (h->prev != e)
	die("BAD LINK");
      e = h;
      h = h->next;
    }
  while (h != &esc_node);
  fputc('\n', stderr);
}

#endif

static inline void
hrelink(struct hnode *o, struct hnode *n)
{
  if (!o->parent)
    hroot = n;
  else if (o->parent->son[0] == o)
    o->parent->son[0] = n;
  else
    o->parent->son[1] = n;
}

static void
hupdate(struct hnode *h)
{
  struct hnode *l, *z, *hp, *ln;
  int rest = 0;

  if (h->parent->son[0] == &esc_node || h->parent->son[1] == &esc_node)
    rest = 1;
  while (h->parent)
    {
      l = z = h;
      while (z->cnt == h->cnt && z != &esc_node) /* Find last node with the same number */
	{
	  if (!rest || z->symbol != INTERNAL)
	    l = z;
	  z = z->next;
	}
      if (l != h)
	{
	  hrelink(l, h);		/* Swap the nodes */
	  hrelink(h, l);
	  z = l->parent;
	  l->parent = h->parent;
	  h->parent = z;
	  h->prev->next = h->next;	/* Unlink h */
	  h->next->prev = h->prev;
	  hp = h->prev;
	  l->prev->next = l->next;	/* Unlink l */
	  l->next->prev = l->prev;
	  ln = l->next;
	  hp->next->prev = l;		/* Link l on h's original place */
	  l->next = hp->next;
	  l->prev = hp;
	  hp->next = l;
	  ln->prev->next = h;		/* Link h on l's original place */
	  h->prev = ln->prev;
	  h->next = ln;
	  ln->prev = h;
	}
      h->cnt++;
      h = h->parent;
      rest = 0;
    }
}

static void
hnew(int sym)
{
  struct hnode *z = &esc_node;
  struct hnode *i = xmalloc(sizeof(struct hnode));
  struct hnode *h = xmalloc(sizeof(struct hnode));

  hrelink(z, i);
  i->parent = z->parent;
  i->son[0] = z;
  i->son[1] = h;
  i->symbol = INTERNAL;
  i->cnt = 0;
  z->parent = i;
  h->parent = i;
  h->son[0] = h->son[1] = NULL;
  h->symbol = sym;
  h->cnt = 0;
  hindex[sym] = h;
  i->next = z->next;
  i->next->prev = i;
  h->next = i;
  i->prev = h;
  z->next = h;
  h->prev = z;
  hupdate(h);
}

/*
 *	Encoder
 */

void
huff_encode(int sym)
{
  struct hnode *h, *p;

  if (!(h = hindex[sym]))
    {
      int i;
      huff_encode(ESC);
      for(i=bits_per_sym-1; i>=0; i--)
	put_bit(sym & (1 << i));
      hnew(sym);
      return;
    }
  p = NULL;
  while (h)
    {
      h->aux = p;
      p = h;
      h = h->parent;
    }
  h = hroot;
  while (h->aux)
    {
      if (h->aux == h->son[0])
	put_bit(0);
      else
	put_bit(1);
      h = h->aux;
    }
  if (sym >= 0)
    hupdate(hindex[sym]);
}

/*
 *	Decoder
 */

int
huff_decode(void)
{
  struct hnode *h = hroot;
  int s, i;

  while (h->symbol == INTERNAL)
    h = h->son[get_bit()];
  s = h->symbol;
  if (s == ESC)
    {
      s = 0;
      for(i=0; i<bits_per_sym; i++)
	s = (s << 1) | get_bit();
      hnew(s);
    }
  else
    hupdate(h);
  return s;
}

#ifdef DEBUG

int
main(int argc, char **argv)
{
  if (argv[1][0] == 'e')
    {
      int i;
      huff_init(stdout, 257);
      while ((i = getchar()) != EOF)
	huff_encode(i);
      huff_encode(256);
      huff_end_write();
    }
  else
    {
      int i;
      huff_init(stdin, 257);
      while ((i = huff_decode()) < 256)
	putchar(i);
    }
  hdump();

  return 0;
}

#endif
